package img;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import javax.swing.ImageIcon;
import java.awt.*;
import java.io.*;
import java.awt.geom.*;

/**
 * This class encapsulates a digital picture, which can be created from a file
 * name, from another image, or from a width and height specification. A digital
 * picture has accessor and mutator methods for its pixels. A digital picture
 * has an associated string which serves as the picture's title.
 * 
 * A digital picture can be displayed graphically. As well, some basic
 * interaction with a digital picture is provided via the PictureExplorer.
 * 
 * This class is based on the Picture class from Guzdial and Ericson, with
 * explanatory comments and modifications provided by various contributors:
 * sdc@cs.albany.edu, ericson@cc.gatech.edu. mb@cse.yorku.ca
 */
public class DigitalPicture implements AbstractDigitalPicture {

	private String fileName;
	private String theTitle;
	private BufferedImage bufferedImage;
	private PictureFrame pictureFrame;
	private String fileNameExtension; // e.g., (jpg or bmp)

	/**
	 * Default constructor, which instantiates a blank picture consisting of
	 * only white pixels, with a default width of 200 and a default height of
	 * 100. The default title is "New Picture".
	 */
	public DigitalPicture() {
		this(200, 100);
	}

	/**
	 * Constructs a picture with the width and height as specified by the passed
	 * parameter values. All of the pixels are set to the default colour of
	 * white. The default title is "New Picture".
	 * 
	 * @param width
	 *            the width of the desired picture
	 * @param height
	 *            the height of the desired picture
	 */
	public DigitalPicture(int width, int height) {
		bufferedImage = new BufferedImage(width, height,
				BufferedImage.TYPE_INT_RGB);
		theTitle = "New Picture";
		fileNameExtension = "jpg";
		setAllPixelsToAColor(Color.white);
	}

	/**
	 * Constructs a picture from the image in the filename that is passed as
	 * parameter. The width and height of the picture will be determined by the
	 * images that is contained in the file. The title will be the name of the
	 * file. If the specified file does not contain an image, then this
	 * constructor causes a runtime exception to occur.
	 * 
	 * 
	 * @param fileName
	 *            the name of the file that contains the source image
	 * @throws Exception
	 *             as described above
	 */
	public DigitalPicture(String fileName) {

		try {
			load(fileName);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Constructs a new instance of this class that has the same state as the
	 * passed instance.
	 * 
	 * @param copyPicture
	 *            the picture to copy from
	 */
	public DigitalPicture(DigitalPicture copyPicture) {
		this.fileName = new String(copyPicture.fileName);
		this.theTitle = new String(copyPicture.theTitle);
		this.fileNameExtension = copyPicture.fileNameExtension;
		if (copyPicture.bufferedImage != null) {
			this.bufferedImage = new BufferedImage(copyPicture.getWidth(),
					copyPicture.getHeight(), BufferedImage.TYPE_INT_RGB);
			this.copyPicture(copyPicture);
		}
	}

	/**
	 * Returns the extension of the filename used to create this picture (e.g.,
	 * jpg or bmp), if one exists (otherwise, null is returned).
	 * 
	 * @return as described above
	 */
	public String getExtension() {
		return fileNameExtension;
	}

	/**
	 * Causes all attributes of this object to be copied from the passed source
	 * picture.
	 * 
	 * @param sourcePicture
	 *            the picture object from which the attributes should be copied.
	 */
	public void copyPicture(DigitalPicture sourcePicture) {
		Pixel sourcePixel = null;
		Pixel targetPixel = null;

		// loop through the columns
		for (int sourceX = 0, targetX = 0; sourceX < sourcePicture.getWidth()
				&& targetX < this.getWidth(); sourceX++, targetX++) {
			// loop through the rows
			for (int sourceY = 0, targetY = 0; sourceY < sourcePicture
					.getHeight() && targetY < this.getHeight(); sourceY++, targetY++) {
				sourcePixel = sourcePicture.getPixel(sourceX, sourceY);
				targetPixel = this.getPixel(targetX, targetY);
				targetPixel.setColor(sourcePixel.getColor());
			}
		}

	}

	/**
	 * Set the color of all of the pixels to the passed color
	 * 
	 * @param color
	 *            the new color for all of the pixels
	 */
	public void setAllPixelsToAColor(Color color) {
		// loop through all x
		for (int x = 0; x < this.getWidth(); x++) {
			// loop through all y
			for (int y = 0; y < this.getHeight(); y++) {
				getPixel(x, y).setColor(color);
			}
		}
	}

	/**
	 * The method causes all of the pixels in the specified row to be changed to
	 * a black colour.
	 * 
	 * @param row
	 */
	public void blacken(int row) {
		for (int i = 0; i < this.getWidth(); i++) {
			Pixel p = getPixel(i, row);
			p.setRed(0);
			p.setGreen(0);
			p.setBlue(0);
			this.repaint();
		}
	}
	
	/**
	 * Cause the pixels in the specified row to change to the specifed colour.
	 * The row indexing is the same as the pixel indexing (e.g., the topmost row
	 * has index position 0)
	 * 
	 * 
	 * @param row
	 *            the index of the row to be changed
	 * @param theNewColor
	 *            the new colour of the row
	 */
	public void modifyRowColour(int row, Color theNewColor) {
		for (int i = 0; i < this.getWidth(); i++) {
			Pixel p = getPixel(i, row);
			p.setColor(theNewColor);
			this.repaint();
		}
	}

	/**
	 * Returns the buffered image associated with this picture.
	 * 
	 * @return as described above
	 */
	public BufferedImage getBufferedImage() {
		return bufferedImage;
	}

	/**
	 * Returns the Graphics instance (encapsulates the graphics context for the
	 * platform upon which the virtual machine is running).
	 * 
	 * @return as described above
	 */
	public Graphics getGraphics() {
		return bufferedImage.getGraphics();
	}

	/**
	 * Returns the Graphics2D instance (encapsulates the graphics context for
	 * the platform upon which the virtual machine is running).
	 * 
	 * @return as described above
	 */
	public Graphics2D getGraphics2D() {
		// mb: did promotion cast here rather than require client to do it
		return (Graphics2D) bufferedImage.getGraphics();
	}

	/**
	 * Returns the Graphics2D instance (encapsulates the graphics context for
	 * the platform upon which the virtual machine is running), which can then
	 * be used to do 2D drawing on the picture
	 * 
	 * @return as described above
	 */
	public Graphics2D createGraphics() {
		return bufferedImage.createGraphics();
	}

	/**
	 * Get the name of the file associated with this digital picture, if one
	 * exists. The digital picture may not have been created from a file, in
	 * which case this method returns null.
	 * 
	 * @return as described above
	 */
	public String getFileName() {
		return fileName;
	}

	/**
	 * Get the string that encapsulates the title that is associated with this
	 * digital picture.
	 * 
	 * @return as described above
	 */
	public String getTitle() {
		return theTitle;
	}

	/**
	 * Set the string that encapsulates the title that is associated with this
	 * digital picture.
	 * 
	 */
	public void setTitle(String title) {
		this.theTitle = title;
		pictureFrame.setTitle(title);
	}

	/**
	 * Returns the width of this picture in pixels.
	 * 
	 * @return as described above.
	 */
	public int getWidth() {
		return bufferedImage.getWidth();
	}

	/**
	 * Returns the height of this picture in pixels.
	 * 
	 * @return as described above.
	 */
	public int getHeight() {
		return bufferedImage.getHeight();
	}

	/**
	 * Returns the PictureFrame object that associated with this picture (it may
	 * be null)
	 * 
	 * @return as described above.
	 */
	public PictureFrame getPictureFrame() {
		return pictureFrame;
	}

	/**
	 * Set the passed PictureFrame object to be associated with this picture.
	 * 
	 * @param pictureFrame
	 *            the picture frame to use
	 */
	public void setPictureFrame(PictureFrame pictureFrame) {
		// set this picture objects' picture frame to the passed one
		this.pictureFrame = pictureFrame;
	}

	/**
	 * Returns the Image from the picture
	 * 
	 * @return as described above.
	 */
	public Image getImage() {
		return bufferedImage;
	}

	/**
	 * Returns the colour information about the specified pixel, as encoded in
	 * an int value.
	 * 
	 * @param x
	 *            the x position of the pixel
	 * @param y
	 *            the y position of the pixel
	 * @return as described above
	 */
	public int getBasicPixel(int x, int y) {
		return bufferedImage.getRGB(x, y);
	}

	/**
	 * Sets the colour of the specified pixel, as encoded in the passed int
	 * value.
	 * 
	 * @param x
	 *            the x position of the pixel
	 * @param y
	 *            the y position of the pixel
	 * @param rgb
	 *            the new rgb value of the pixel (alpha, red, green, blue)
	 */
	public void setBasicPixel(int x, int y, int rgb) {
		bufferedImage.setRGB(x, y, rgb);
	}

	/**
	 * Returns the Pixel object at the specified (x,y) coordinate.
	 * 
	 * @param x
	 *            the x position of the pixel
	 * @param y
	 *            the y position of the pixel
	 * @return as descibed above
	 */
	public Pixel getPixel(int x, int y) {
		// create the pixel object for this picture and the given x and y
		// location
		Pixel pixel = new Pixel(this, x, y);
		return pixel;
	}

	/**
	 * Returns a one-dimensional array of Pixels for this picture
	 * 
	 * @return a one-dimensional array of Pixel objects starting with y=0 to
	 *         y=height-1 and x=0 to x=width-1.
	 */
	public Pixel[] getPixels() {
		int width = getWidth();
		int height = getHeight();
		Pixel[] pixelArray = new Pixel[width * height];

		// loop through height rows from top to bottom
		for (int row = 0; row < height; row++)
			for (int col = 0; col < width; col++)
				pixelArray[row * width + col] = new Pixel(this, col, row);

		return pixelArray;
	}

	/**
	 * Load the buffered image with the passed image
	 * 
	 * @param image
	 *            the image to use
	 */
	public void load(Image image) {
		// get a graphics context to use to draw on the buffered image
		Graphics2D graphics2d = bufferedImage.createGraphics();

		// draw the image on the buffered image starting at 0,0
		graphics2d.drawImage(image, 0, 0, null);

		// show the new image
		show();
	}

	/**
	 * Causes this picture to be displayed graphically (using the PictureFrame
	 * object that is associated with this picture).
	 */
	public void show() {
		// if there is a current picture frame then use it
		if (pictureFrame != null)
			pictureFrame.updateImageAndShowIt();

		// else create a new picture frame with this picture
		else
			pictureFrame = new PictureFrame(this);
	}

	/**
	 * Causes this picture to be shown in a new PictureExplorer window.
	 */
	public void explore() {
		// create a copy of the current picture and explore it
		new PictureExplorer(new DigitalPicture(this));
	}

	/**
	 * Forces this picture to redraw itself. Necessary after the pixels in a
	 * picture have been changed.
	 */
	public void repaint() {
		// if there is a picture frame tell it to repaint
		if (pictureFrame != null)
			pictureFrame.repaint();

		// else create a new picture frame
		else
			pictureFrame = new PictureFrame(this);
	}

	/**
	 * Loads the picture from the passed file name, if a picture exists in the
	 * file. If not, this method throws an exception.
	 * 
	 * @param fileName
	 *            the file name to use to load the picture from
	 * @throws Exception
	 *             as described above
	 */
	public void load(String fileName) throws Exception {

		// set the current piture's file name
		this.fileName = fileName;

		// set the extension
		int posDot = fileName.indexOf('.');
		if (posDot >= 0)
			this.fileNameExtension = fileName.substring(posDot + 1);

		// if the current title is null use the file name
		if (theTitle == null)
			theTitle = fileName;

		// try to get the buffered image from the file
		try {
			bufferedImage = ImageIO.read(new File(this.fileName));
			if (bufferedImage == null)
				throw new Exception("null BufferedImage");
			/*
			 * if anything goes wrong tell the user on the console, set the
			 * image to a black image with a message on it, and return false
			 */
		} catch (Exception ex) {
			String MSG = "Couldn't load the file " + this.fileName;
			throw new Exception(MSG);
		}
	}

	/**
	 * Draw the passed string on this picture at the specified coordinates. The
	 * default font is Helvetica, bold, with point size 16.
	 * 
	 * @param message
	 *            the message to draw on the buffered image
	 * @param xPos
	 *            the leftmost point of the string in x
	 * @param yPos
	 *            the bottom of the string in y
	 */
	public void addMessage(String message, int xPos, int yPos) {
		// get a graphics context to use to draw on the buffered image
		Graphics2D graphics2d = bufferedImage.createGraphics();

		// set the color to white
		graphics2d.setPaint(Color.white);

		// set the font to Helvetica bold style and size 16
		graphics2d.setFont(new Font("Helvetica", Font.BOLD, 16));

		// draw the message
		graphics2d.drawString(message, xPos, yPos);

	}

	/**
	 * Method to create a new picture of the passed height. The aspect ratio of
	 * the width and height will stay the same.
	 * 
	 * @param height
	 *            the desired height
	 * @return the resulting picture
	 */
	public DigitalPicture getPictureWithHeight(int height) {
		// set up the scale tranform
		double yFactor = (double) height / this.getHeight();
		AffineTransform scaleTransform = new AffineTransform();
		scaleTransform.scale(yFactor, yFactor);

		// create a new picture object that is the right size
		DigitalPicture result = new DigitalPicture(
				(int) (getWidth() * yFactor), (int) (getHeight() * yFactor));

		// get the graphics 2d object to draw on the result
		Graphics graphics = result.getGraphics();
		Graphics2D g2 = (Graphics2D) graphics;

		// draw the current image onto the result image scaled
		g2.drawImage(this.getImage(), scaleTransform, null);

		return result;
	}

	/**
	 * Method to write the contents of the picture to a file with the passed
	 * name
	 * 
	 * @param fileName
	 *            the name of the file to write the picture to
	 */
	public void write(String fileName) {
		String extension = this.fileNameExtension; // the default is current

		try {
			// create the file object
			File file = new File(fileName);

			// get the extension
			int posDot = fileName.indexOf('.');
			if (posDot >= 0)
				extension = fileName.substring(posDot + 1);

			// write the contents of the buffered image to the file as jpeg
			ImageIO.write(bufferedImage, extension, file);

		} catch (Exception ex) {
			System.out.println("Couldn't write out the picture to the file "
					+ fileName);
		}

	}

	/**
	 * Method to get the coordinates of the enclosing rectangle after this
	 * transformation is applied to the current picture
	 * 
	 * @return the enclosing rectangle
	 */
	public Rectangle2D getTranslationEnclosingRectangle(AffineTransform trans) {
		int width = getWidth();
		int height = getHeight();
		double maxX = width - 1;
		double maxY = height - 1;
		double minX, minY;
		Point2D.Double p1 = new Point2D.Double(0, 0);
		Point2D.Double p2 = new Point2D.Double(maxX, 0);
		Point2D.Double p3 = new Point2D.Double(maxX, maxY);
		Point2D.Double p4 = new Point2D.Double(0, maxY);
		Point2D.Double result = new Point2D.Double(0, 0);
		Rectangle2D.Double rect = null;

		// get the new points and min x and y and max x and y
		trans.deltaTransform(p1, result);
		minX = result.getX();
		maxX = result.getX();
		minY = result.getY();
		maxY = result.getY();
		trans.deltaTransform(p2, result);
		minX = Math.min(minX, result.getX());
		maxX = Math.max(maxX, result.getX());
		minY = Math.min(minY, result.getY());
		maxY = Math.max(maxY, result.getY());
		trans.deltaTransform(p3, result);
		minX = Math.min(minX, result.getX());
		maxX = Math.max(maxX, result.getX());
		minY = Math.min(minY, result.getY());
		maxY = Math.max(maxY, result.getY());
		trans.deltaTransform(p4, result);
		minX = Math.min(minX, result.getX());
		maxX = Math.max(maxX, result.getX());
		minY = Math.min(minY, result.getY());
		maxY = Math.max(maxY, result.getY());

		// create the bounding rectangle to return
		rect = new Rectangle2D.Double(minX, minY, maxX - minX + 1, maxY - minY
				+ 1);
		return rect;
	}

	/**
	 * Return a string representation of this Picture. The string contains the
	 * height and width of the image.
	 * 
	 * 
	 * @return as described above.
	 */
	public String toString() {
		String output = "Picture, filename " + getFileName() + " height "
				+ getHeight() + " width " + getWidth();
		return output;

	}

}